;
;PRNG16 is unoptimised PRNG16 routine
;
;On entry, BCDE = last state
;On exit,  BCDE = new  state, HL = new pseudo-random number
;
;Destroys AF
;
;Define X = P XOR Q, then
;
; P := P ROL 14 XOR X SHL 2 XOR X
; Q := X ROL 7
;
;which is equivalent to
;
; P := P ROR 2 XOR X SHL 2 XOR X
; Q := X ROL 8 ROR 1
;
PRNG16:
;
;DE = Pf Pe Pd Pc Pb Pa P9 P8  P7 P6 P5 P4 P3 P2 P1 P0
;BC = Qf Qe Qd Qc Qb Qa Q9 Q8  Q7 Q6 Q5 Q4 Q3 Q2 Q1 Q0
;
	LD A,C
	XOR E
	LD L,A
	LD A,B
	XOR D
	LD H,A			;HL = X
;
	LD B,H
	LD C,L			;BC = X
;
	ADD HL,HL
	ADD HL,HL		;HL = X SLA 2
;
	LD A,E
	RRA
	RR D
	RR E
	RRA
	RR D
	RR E			;DE = P ROR 2
;
	LD A,E
	XOR C
	XOR L
	LD E,A
	LD A,D
	XOR B
	XOR H
	LD D,A			;DE = P ROR 2 XOR X SHL 2 XOR X = new P
;
	LD A,C
	LD C,B
	LD B,A			;BC = X ROL 8
	RRA
	RR B
	RR C			;BC = X ROL 8 ROR 1 = X ROL 7 = new Q
;
	LD H,B
	LD L,C
	ADD HL,DE		;HL = new PRN
	RET
;
;	40 bytes excluding RET
;
;
;
;BEST16 is the best optimised PRNG16 routine (by TB)
;
;On entry, BCDE = last state
;On exit,  BCDE = new  state, HL = pseudo-random number
;
;Destroys AF
;
BEST16:
;
;DE = Pf Pe Pd Pc Pb Pa P9 P8  P7 P6 P5 P4 P3 P2 P1 P0
;BC = Qf Qe Qd Qc Qb Qa Q9 Q8  Q7 Q6 Q5 Q4 Q3 Q2 Q1 Q0
;
	LD A,C
	XOR E
	LD L,A
	LD A,B
	XOR D
	LD H,A			;HL = X
;
	LD C,H
	LD B,L			;CB = X so BC = X ROL 8
;
	ADD HL,HL
	ADD HL,HL		;HL = X SLA 2
;
	LD A,E
	RR E
	RR D
	RRA
	RR E
	RR D
	RRA			;DA = P ROR 2
;
;	LD A,E			;*** byte saved ***
	XOR B
	XOR L
	LD E,A
	LD A,D
	XOR C
	XOR H
	LD D,A			;DE = P ROR 2 XOR X SHL 2 XOR X = new P
;
	LD A,C
;	LD C,B			;*** byte saved ***
;	LD B,A			;*** byte saved ***
	RRA
	RR B
	RR C			;BC = X ROL 8 ROR 1 = X ROL 7 = new Q
;
	LD H,B
	LD L,C
	ADD HL,DE		;HL = new PRN
	RET
;
;	37 bytes excluding RET
;
;
;
;BARE16 is BEST16 with comments removed and timings added
;
BARE16:
	LD A,C			; 4
	XOR E			; 4
	LD L,A			; 4
	LD A,B			; 4
	XOR D			; 4
	LD H,A			; 4
	LD C,H			; 4
	LD B,L			; 4
	ADD HL,HL		;11
	ADD HL,HL		;11
	LD A,E			; 4
	RR E			; 8
	RR D			; 8
	RRA			; 4
	RR E			; 4
	RR D			; 8
	RRA			; 8
	XOR B			; 4
	XOR L			; 4
	LD E,A			; 4
	LD A,D			; 4
	XOR C			; 4
	XOR H			; 4
	LD D,A			; 4
	LD A,C			; 4
	RRA			; 4
	RR B			; 8
	RR C			; 8
	LD H,B			; 4
	LD L,C			; 4
	ADD HL,DE		;11
	RET
;
;	22 *  4T   88T	22 bytes
;	 6 *  8T   48T	12 bytes
;	 3 * 11T   33T	 3 bytes
;		  ----	--------
;		  169T	37 bytes
;
;
;
;BEST17 is best optimised PRNG17 routine (by TB)
;
;On entry, BCDEA = last state
;On exit,  BCDEA = new  state, HL = new pseudo-random number
;
;Destroys F
;
;Define X = P XOR Q, then
;
; P := P ROL 7 XOR X SHL 8 XOR X
; Q := X ROL 12
;
;which is equivalent to
;
; P := P ROL 8 ROR 1 XOR X SHL 8 XOR X
; Q := X ROL 8 ROL 4
;
BEST17:
;
;DE = Pg Pf Pe Pd Pc Pb Pa P9  P8 P7 P6 P5 P4 P3 P2 P1
;BC = Qg Qf Qe Qd Qc Qb Qa Q9  Q8 Q7 Q6 Q5 Q4 Q3 Q2 Q1
;A  = P0 X0 0  0  0  0  0  0
;
	LD H,A			;Save A in H
;
	LD A,C
	XOR E
	LD C,A
	LD A,B
	XOR D			;CF = 0
	LD B,A
;
;DE = Pg Pf Pe Pd Pc Pb Pa P9  P8 P7 P6 P5 P4 P3 P2 P1
;BC = Xg Xf Xe Xd Xc Xb Xa X9  X8 X7 X6 X5 X4 X3 X2 X1
;HL = P0 X0 0  0  0  0  0  0   ?  ?  ?  ?  ?  ?  ?  ?
;CF = 0
;
;P ROL 7 = P ROL 8 ROR 1
;	 = P9 P8 P7 P6 P5 P4 P3 P2  P1 P0 Pg Pf Pe Pd Pc Pb  Pa
;
	RL H			;H = X0 0  0  0  0  0  0  0  | P0 -> CF   (RL safe as CF = 0)
	LD A,D			;A = Pg Pf Pe Pd Pc Pb Pa P9
	RRA			;A = P0 Pg Pf Pe Pd Pc Pb Pa | P9 -> CF
	RR E			;E = P9 P8 P7 P6 P5 P4 P3 P2 | P1 -> CF
	RRA			;A = P1 P0 Pg Pf Pe Pd Pc Pb | Pa -> CF
	LD D,E			;D = P9 P8 P7 P6 P5 P4 P3 P2
;
;DA = P9 P8 P7 P6 P5 P4 P3 P2  P1 P0 Pg Pf Pe Pd Pc Pb = P ROL 7 [16:1] 
;BC = Xg Xf Xe Xd Xc Xb Xa X9  X8 X7 X6 X5 X4 X3 X2 X1
;HL = X0 0  0  0  0  0  0  0   ?  ?  ?  ?  ?  ?  ?  ?
;CF = Pa
;
	LD L,H			;L = X0 0  0  0  0  0  0  0
	RL L			;L = 0  0  0  0  0  0  0  Pa | X0 -> CF (discard)
;
;HL = X0 0  0  0  0  0  0  0   0  0  0  0  0  0  0  Pa
;DA = P9 P8 P7 P6 P5 P4 P3 P2  P1 P0 Pg Pf Pe Pd Pc Pb = P ROL 7 [16:1]
;BC = Xg Xf Xe Xd Xc Xb Xa X9  X8 X7 X6 X5 X4 X3 X2 X1 = X       [16:1]
;CH = X8 X7 X6 X5 X4 X3 X2 X1  X0 0  0  0  0  0  0  0  = X SHL 8 [16:1]
;
;P := P ROL 7 XOR X XOR X SHL 8
;P[16:1} := DA XOR BC XOR CH -> DE
;P[8:1]  :=  A  XOR C XOR  H ->  E
;P[16:9] := D  XOR B  XOR C  -> D
;
	XOR C
	XOR H
	LD E,A
	LD A,D
	XOR B
	XOR C
	LD D,A
;
;DE = Pg Pf Pe Pd Pc Pb Pa P9  P8 P7 P6 P5 P4 P3 P2 P1 = new P[16:1]
;BC = Xg Xf Xe Xd Xc Xb Xa X9  X8 X7 X6 X5 X4 X3 X2 X1
;HL = X0 0  0  0  0  0  0  0   0  0  0  0  0  0  0  Pa
;
;Q := X ROL 12 = X ROL 8 ROL 4
;	       = X4 X3 X2 X1 X0 Xg Xf Xe  Xd Xc Xb Xa X9 X8 X7 X6  X5
;
	RLC H			;H = 0  0  0  0  0  0  0  X0 | X0 -> CF and keep X0 in H
	LD A,C			;A = X8 X7 X6 X5 X4 X3 X2 X1
	RLA			;A = X7 X6 X5 X4 X3 X2 X1 X0 | X8 -> CF
	RL B			;B = Xf Xe Xd Xc Xb Xa X9 X8 | Xg -> CF
	RLA			;A = X6 X5 X4 X3 X2 X1 X0 Xg | X7 -> CF
	RL B			;B = Xe Xd Xc Xb Xa X9 X8 X7 | Xf -> CF
	RLA			;A = X5 X4 X3 X2 X1 X0 Xg Xf | X6 -> CF
	RL B			;B = Xd Xc Xb Xa X9 X8 X7 X6 | Xe -> CF
	RLA			;A = X4 X3 X2 X1 X0 Xg Xf Xe | X5 -> CF
	LD C,B			;C = Xd Xc Xb Xa X9 X8 X7 X6
	LD B,A			;B = X4 X3 X2 X1 X0 Xg Xf Xe
;
;DE = Pg Pf Pe Pd Pc Pb Pa P9  P8 P7 P6 P5 P4 P3 P2 P1 = new P[16:1]
;BC = X4 X3 X2 X1 X0 Xg Xf Xe  Xd Xc Xb Xa X9 X8 X7 X6 = new Q[16:1]
;HL = 0  0  0  0  0  0  0  X0  0  0  0  0  0  0  0  Pa
;CF = X5 = new Q0
;
;P0 := (P ROL 7 XOR X XOR X SHL 8)[0] = Pa XOR X0 XOR 0
;
	LD A,L			;A = 0  0  0  0  0  0  0  Pa
	RR L			;L = Q0 0  0  0  0  0  0  0  | Pa -> CF (discard)   Q0 = new Q0
	XOR H			;A = 0  0  0  0  0  0  0  P0 | 0  -> CF		    P0 = new P0
	RRCA			;A = P0 0  0  0  0  0  0  0  | P0 -> CF
	ADC A,L			;A = X0 0  0  0  0  0  0  P0 | P0+Q0 carry -> CF    X0 = new X0
;
;ADC A,L has three effects:
;
; (1) Bit 0 of A = carry input = P0
; (2) Bit 7 of A = P0 XOR Q0 = X0
; (3) Carry output = P0 AND Q0
;
	LD H,B
	LD L,C
	ADC HL,DE		;HL = new PRN
	RRCA			;A = P0 X0 0  0  0  0  0  0 | P0 -> CF (discard) 
	RET
;
;	51 bytes excluding RET
;
;
;
;BARE17 is BEST17 with comments removed and timings added
;
BARE17:
	LD H,A			; 4
	LD A,C			; 4
	XOR E			; 4
	LD C,A			; 4
	LD A,B			; 4
	XOR D			; 4
	LD B,A			; 4
	RL H			; 8
	LD A,D			; 4
	RRA			; 4
	RR E			; 8
	RRA			; 4
	LD D,E			; 4
	LD L,H			; 4
	RL L			; 8
	XOR C			; 4
	XOR H			; 4
	LD E,A			; 4
	LD A,D			; 4
	XOR B			; 4
	XOR C			; 4
	LD D,A			; 4
	RLC H			; 8
	LD A,C			; 4
	RLA			; 4
	RL B			; 8
	RLA			; 4
	RL B			; 8
	RLA			; 4
	RL B			; 8
	RLA			; 4
	LD C,B			; 4
	LD B,A			; 4
	LD A,L			; 4
	RR L			; 8
	XOR H			; 4
	RRCA			; 4
	ADC A,L			; 4
	LD H,B			; 4
	LD L,C			; 4
	ADC HL,DE		;15
	RRCA			; 4
	RET
;
;	33 *  4T  132T	33 bytes
;	 8 *  8T   64T  16 bytes
;	 1 * 15T   15T   2 bytes
;		  ----  --
;		  211T  51 bytes
;
